defmodule Profile do
  def start(name) do
    Process.put({name, :start}, now_ms())
  end

  def stop(name) do
    start = Process.get({name, :start}, 0)
    exec_time = now_ms() - start
    sum = Process.get({name, :sum}, 0)
    Process.put({name, :sum}, sum + exec_time)
  end

  def flush(name) do
    Process.delete({name, :sum})
  end

  defmacro flame({{:., _, [module, function]}, _, args}) do
    quote bind_quoted: [module: module, function: function, args: args],
          do: :eflame.apply(:normal, Profile.filename(), module, function, args)
  end

  def filename() do
    pid_str = self() |> :erlang.pid_to_list() |> to_string()
    pid_name = String.slice(pid_str, 1, byte_size(pid_str) - 2)
    uniq_int = :erlang.unique_integer([:positive, :monotonic])
    "stacks_#{pid_name}_#{uniq_int}_.out"
  end

  def init(_) do
    {:ok, %{}}
  end

  defmacro timed(name, block, flush \\ []) do
    quote do
      start = :erlang.system_time(:microsecond)
      result = unquote(block)
      Profile.print(unquote(name), start, :erlang.system_time(:microsecond), unquote(flush))
      result
    end
  end

  @pattern Logger.Formatter.compile("$time $metadata $message")
  def print(name, start, stop, flush) do
    results =
      Enum.map_join(flush, "; ", fn name -> "#{name}: #{name |> flush() |> time_used()}" end)

    message = "#{name} #{time_used(stop - start)} (#{results})"
    formatted = Logger.Formatter.format(@pattern, :info, message, timestamp(start), pid: self())
    formatted |> IO.chardata_to_string() |> IO.puts()
  end

  def log(message) do
    start = :erlang.system_time(:microsecond)
    formatted = Logger.Formatter.format(@pattern, :info, message, timestamp(start), pid: self())
    formatted |> IO.chardata_to_string() |> IO.puts()
  end

  def now_ms(), do: :erlang.system_time(:microsecond)

  def now() do
    timestamp(now_ms())
  end

  def timestamp(ts) do
    {date, time} = :calendar.system_time_to_local_time(ts, :microsecond)
    ms = ts |> div(1000) |> rem(1000)
    {date, Tuple.insert_at(time, 3, ms)}
  end

  def time_used(time) when time < 1000, do: time
  def time_used(time) when time < 2_000_000, do: "#{div(time, 1000)}ms"
  def time_used(time), do: "#{div(time, 1_000_000)}s"
end
